; Ranrot macros & data
;
; These are high-quality high-speed uniform random number generators.
;
; Thanks to:
;  - Agner Fog for the original code, on which these macros are based (http://www.agner.org/random/)
;  - Mark Larson for his ideas on how to speed up the code
;

; Equates for the generator
Ranrot_JJ	equ	10	; Lag
Ranrot_KK	equ	17	; Circular buffer size
Ranrot_R1	equ	19	; Rotate count A
Ranrot_R2	equ	27	; Rotate count B

; Generates a 64-bit random number and stores it in EDX:EAX
;
; Uses: eax, ebx, ecx, edx
WBRandom MACRO
	mov ebx, Ranrot_p1
	mov ecx, Ranrot_p2
	mov edx, Ranrot_buf[ebx]
	mov eax, Ranrot_buf[ebx+4]
	
	rol edx, Ranrot_R1
	rol edx, Ranrot_R2
	add edx, Ranrot_buf[ecx]
	add eax, Ranrot_buf[ecx+4]
	
	mov Ranrot_buf[ebx], eax
	mov Ranrot_buf[ebx+4], edx
	
	mov ebx, Ranrot_next[ebx]
	mov ecx, Ranrot_next[ecx]
	
	mov Ranrot_p1, ebx
	mov Ranrot_p2, ecx
ENDM

; Generates a 32-bit number and stores it in EAX
;
; Uses: eax, ebx, ecx, edx
WBRandom2 MACRO
	LOCAL skip
	LOCAL calc
	
	mov eax, Ranrot_last
	test eax, eax
	jz calc
	
	xor edx, edx
	mov Ranrot_last, edx
	jmp skip
	
calc:
	WBRandom
	
	mov Ranrot_last, edx
skip:
ENDM

; Generates a double-precision floating point number in xmm0
;
; Uses: eax, ebx, ecx, edx, xmm0
WDRandom MACRO

	WBRandom
	
	and edx, 000FFFFFh
	or edx, 3FF00000h
	
	mov dword ptr Ranrot_double, eax
	mov dword ptr Ranrot_double+4, edx
	
	movlpd xmm0, Ranrot_double
	subsd xmm0, Ranrot_double_one
ENDM

; Generates a single-precision floating point number in xmm0
;
; Uses: eax, ebx, ecx, edx, xmm0
WSRandom MACRO

	WBRandom2
	
	and eax, 007FFFFFh
	or eax, 3F800000h
	movd xmm0, eax
	subss xmm0, qword ptr Ranrot_single_one
ENDM

; Generates an integer in the interval [min,max[ and stores it in eax
;
; Uses: eax, ebx, ecx, edx
; Note: min and max must be constants
WIRandomC MACRO min:req, max:req

	WBRandom
	
	mov ebx, max-min
	mov ecx, edx
	mul ebx
	mov eax, ecx
	mov ecx, edx
	mul ebx
	add eax, ecx
	adc edx, min
	mov eax, edx

ENDM

; Generates an integer in the interval [min,max[ and stores it in eax
;
; Uses: eax, ebx, ecx, edx
; Note: min and max must not be eax, ebx, ecx or edx
WIRandom MACRO min:req, max:req

	WBRandom

	mov ebx, max
	sub ebx, min
	mov ecx, edx
	mul ebx
	mov eax, ecx
	mov ecx, edx
	mul ebx
	add eax, ecx
	adc edx, min
	mov eax, edx
	
ENDM

; Initializes the generator
;
; Uses: eax, ebx, ecx, edx
WRandomInit MACRO seed:req
	local _loop

	mov eax, seed
	xor ecx, ecx
_loop:
	imul eax, 2891336453
	add eax, 1
	mov Ranrot_buf[4*ecx], eax
	inc ecx
	cmp ecx, Ranrot_KK*2
	jb _loop
	
	mov Ranrot_p1, 0
	mov Ranrot_p2, Ranrot_JJ*8
	
	WBRandom
	
	push edi
	mov edi, 30
_loop2:
	WBRandom
	sub edi, 1
	jnz _loop2
	
	pop edi
ENDM

; Some data for the generators
.data

align 16
Ranrot_p1		DD	0							; First pointer in circular buffer
Ranrot_p2		DD	0							; Second pointer in circular buffer
Ranrot_buf	DD	(2*Ranrot_KK) dup (0); Circular buffer

align 16
Ranrot_double			REAL8	1.0, 1.0
Ranrot_double_one	REAL8	1.0, 1.0

Ranrot_single			REAL4 1.0, 1.0, 1.0, 1.0
Ranrot_single_one	REAL4	1.0, 1.0, 1.0, 1.0

align 16
Ranrot_next	DD	2 dup (8*(Ranrot_KK-1))
						
						counter = 0
						REPEAT (Ranrot_KK-1)*2
							dd 4*counter
							counter = counter + 1
						ENDM
						
Ranrot_last	DD	0
